# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
from enum import Enum, auto
from typing import Iterator, Optional, Pattern, Sequence, Set, Union

import capstone
import gtirb
import gtirb_functions

from .utils import _nonterminator_instructions

ENTRYPOINT_NAME = "\0"
"""
A meta-name that refers to the module's entrypoint.
"""


MAIN_NAME = "\0\0"
"""
A meta-name that refers to the module's main function. This may not always be
called "main" (e.g. on Windows).
"""


class BlockPosition(Enum):
    ENTRY = auto()
    """
    Insert the patch at the beginning of a block.
    """

    EXIT = auto()
    """
    Insert the patch at the end of the block, but before the terminator
    instruction (if there is one).
    """

    ANYWHERE = auto()
    """
    Insert the patch anywhere in the block (except after the terminator),
    allowing bubbling to occur.
    """


class Scope:
    """
    Scopes provide a declarative way to specify where in a program a patch
    should be applied.
    """

    def _function_matches(
        self, module: gtirb.Module, func: gtirb_functions.Function
    ) -> bool:
        """
        Determines if a function matches the scope. If this returns False, no
        patches in this scope will be appled to this function.
        """
        raise NotImplementedError

    def _block_matches(
        self,
        module: gtirb.Module,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
    ) -> bool:
        """
        Determines if a block matches the scope. If this returns False, no
        patches in this scope will be applied to this block.
        """
        raise NotImplementedError

    def _needs_disassembly(self) -> bool:
        """
        Does this scope's _potential_offsets method need the disassembly
        parameter filled out?
        """
        raise NotImplementedError

    def _replacement_length(self) -> int:
        return 0

    def _potential_offsets(
        self,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        disassembly: Optional[Sequence[capstone.CsInsn]],
    ) -> Iterator[int]:
        """
        Returns an iterator of all the potential offsets in the block where
        the patch could be applied.

        Note that only one offset will be chosen, so it is not possible to
        have a scope that inserts more than once in a block.
        """
        raise NotImplementedError


class AllBlocksScope(Scope):
    """
    Specifies that an insertion should happen in all blocks of a program. The
    functions that this inserts into can be controlled with a list of names
    (or regular expressions) to check the name against.
    """

    def __init__(
        self,
        position: BlockPosition,
        exclude_functions: Optional[Set[Union[str, Pattern]]] = None,
    ):
        self.position = position
        self.exclude_functions = exclude_functions

    def _function_matches(
        self, module: gtirb.Module, func: gtirb_functions.Function
    ) -> bool:
        return self.exclude_functions is None or not pattern_match(
            module, func, self.exclude_functions
        )

    def _block_matches(
        self,
        module: gtirb.Module,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
    ) -> bool:
        return True

    def _needs_disassembly(self) -> bool:
        return self.position in [BlockPosition.ANYWHERE, BlockPosition.EXIT]

    def _potential_offsets(
        self,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        disassembly: Optional[Sequence[capstone.CsInsn]],
    ) -> Iterator[int]:
        return _potential_offsets_in_block(self.position, block, disassembly)


class SingleBlockScope(Scope):
    """
    Specifies that an insertion should happen in a specific block of a program.
    """

    def __init__(self, block: gtirb.CodeBlock, position: BlockPosition):
        self.block = block
        self.position = position

    def _function_matches(
        self, module: gtirb.Module, func: gtirb_functions.Function
    ) -> bool:
        return self.block in func.get_all_blocks()

    def _block_matches(
        self,
        module: gtirb.Module,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
    ) -> bool:
        return self.block == block

    def _needs_disassembly(self) -> bool:
        return self.position in [BlockPosition.ANYWHERE, BlockPosition.EXIT]

    def _potential_offsets(
        self,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        disassembly: Optional[Sequence[capstone.CsInsn]],
    ) -> Iterator[int]:
        return _potential_offsets_in_block(self.position, block, disassembly)


class FunctionPosition(Enum):
    ENTRY = auto()
    """
    Insert into function entry blocks.
    """

    EXIT = auto()
    """
    Insert into function exit blocks.
    """


class AllFunctionsScope(Scope):
    """
    Specifies that an insertion should happen either in the entry blocks or
    exit blocks of functions. The functions that this inserts into can be
    controlled with a list of names (or regular expressions) to check the name
    against.
    """

    def __init__(
        self,
        position: FunctionPosition,
        block_position: BlockPosition,
        functions: Set[Union[str, Pattern]] = None,
    ):
        self.position = position
        self.block_position = block_position
        self.functions = functions

    def _function_matches(
        self, module: gtirb.Module, func: gtirb_functions.Function
    ) -> bool:
        return self.functions is None or pattern_match(
            module, func, self.functions
        )

    def _block_matches(
        self,
        module: gtirb.Module,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
    ) -> bool:
        if self.position == FunctionPosition.ENTRY:
            return block in func.get_entry_blocks()
        if self.position == FunctionPosition.EXIT:
            return block in func.get_exit_blocks()
        assert False, f"Invalid position: {self.position}"

    def _needs_disassembly(self) -> bool:
        return self.block_position in [
            BlockPosition.ANYWHERE,
            BlockPosition.EXIT,
        ]

    def _potential_offsets(
        self,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        disassembly: Optional[Sequence[capstone.CsInsn]],
    ) -> Iterator[int]:
        return _potential_offsets_in_block(
            self.block_position, block, disassembly
        )


class _SpecificLocationScope(Scope):
    def __init__(
        self,
        function: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        offset: int,
        replacement_length: int = 0,
    ):
        self.function = function
        self.block = block
        self.offset = offset
        self.replacement_length = replacement_length

    def _function_matches(
        self, module: gtirb.Module, func: gtirb_functions.Function
    ) -> bool:
        return self.function == func

    def _block_matches(
        self,
        module: gtirb.Module,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
    ) -> bool:
        return self.block == block

    def _needs_disassembly(self) -> bool:
        return False

    def _replacement_length(self) -> int:
        return self.replacement_length

    def _potential_offsets(
        self,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        disassembly: Optional[Sequence[capstone.CsInsn]],
    ) -> Iterator[int]:
        yield self.offset


def _potential_offsets_in_block(
    block_position: BlockPosition,
    block: gtirb.CodeBlock,
    disassembly: Optional[Sequence[capstone.CsInsn]],
) -> Iterator[int]:
    if block_position == BlockPosition.ENTRY:
        yield 0
    elif block_position == BlockPosition.ANYWHERE:
        assert disassembly
        offset = 0
        for inst in _nonterminator_instructions(block, disassembly):
            yield offset
            offset += inst.size
        yield offset
    elif block_position == BlockPosition.EXIT:
        assert disassembly
        yield sum(
            inst.size
            for inst in _nonterminator_instructions(block, disassembly)
        )
    else:
        assert False, "Invalid block position"


def pattern_match(
    module: gtirb.Module,
    func: gtirb_functions.Function,
    match_set: Set[Union[str, Pattern]],
) -> bool:
    """
    Determines if a function matches a set of regex patterns or literal names.

    :param module: The gtirb Module containing the Function.
    :param func: The Function to check.
    :param match_set: The names or patterns to match against.
    """
    for fname in match_set:
        if fname == MAIN_NAME:
            # TODO: Main isn't necessarily "main" on Windows.
            matches = func.get_name() == "main"
        elif fname == ENTRYPOINT_NAME:
            matches = module.entry_point in func.get_entry_blocks()
        elif isinstance(fname, Pattern):
            matches = fname.fullmatch(func.get_name()) is not None
        else:
            matches = func.get_name() == fname
        if matches:
            return True
    return False



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import dataclasses
import itertools
import logging
import operator
import uuid
from typing import List, NamedTuple, Sequence, Tuple

import gtirb
import gtirb_functions
import mcasm
from gtirb_capstone.instructions import GtirbInstructionDecoder

from .assembler import _Assembler
from .assembly import InsertionContext, Patch
from .isa import _get_isa
from .prepare import prepare_for_rewriting
from .scopes import Scope, _SpecificLocationScope
from .utils import (
    _modify_block_insert,
    _text_section_name,
    decorate_extern_symbol,
    show_block_asm,
)


class _Insertion(NamedTuple):
    scope: Scope
    patch: Patch


class _FunctionInsertion(NamedTuple):
    symbol: gtirb.Symbol
    block: gtirb.CodeBlock
    patch: Patch


class RewritingContext:
    """
    A rewriting context manages insertions and modifications on a single
    module. It takes care of resolving insertion scopes to concrete positions,
    potentially trying to bubble multiple insertions together for performance.
    """

    def __init__(
        self,
        module: gtirb.Module,
        functions: Sequence[gtirb_functions.Function],
        logger=logging.getLogger("gtirb_rewriting"),
        expensive_assertions=True,
    ):
        """
        :param module: The module to rewrite.
        :param functions: The list of functions in the module.
        :param logger: The logger to log to when rewriting.
        :param expensive_assertions: If enabled, extra assertions will be
        enabled that may have noticable run-time overhead.
        """
        self._module = module
        self._symbols_by_name = {s.name: s for s in module.symbols}
        self._functions = functions
        self._decoder = GtirbInstructionDecoder(self._module.isa)
        self._isa = _get_isa(module)
        self._insertions: List[_Insertion] = []
        self._function_insertions: List[_FunctionInsertion] = []
        self._logger = logger
        self._patch_id = 0
        self._expensive_assertions = expensive_assertions
        self._leaf_functions = {
            f.uuid: self._might_be_leaf_function(f) for f in self._functions
        }

    def _might_be_leaf_function(self, func: gtirb_functions.Function) -> bool:
        """
        Determines if a function might be a leaf function by the absence of
        calls in the CFG. This should only be used _before_ applying rewrites
        because the status may change if a patch inserts a call.
        """
        return all(
            not edge.label or edge.label.type != gtirb.Edge.Type.Call
            for block in func.get_all_blocks()
            for edge in block.outgoing_edges
        )

    def _log_patch_error(
        self,
        asm: str,
        patch: Patch,
        patch_id: int,
        err: mcasm.assembler.AsmSyntaxError,
    ) -> None:
        """
        Logs an assembly syntax error to our logger.
        """
        lines = asm.splitlines()
        self._logger.error("error in %s (#%i): %s", patch, self._patch_id, err)
        for line in lines[: err.lineno]:
            self._logger.error("%s", line)
        # LLVM only stores the start column in its diagnostic object, so we'll
        # highlight the whole rest of the line as the error.
        self._logger.error(
            " " * err.column + "^" + "~" * (len(line) - err.column - 1)
        )
        for line in lines[err.lineno :]:
            self._logger.error("%s", line)

    def _invoke_patch(
        self,
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        offset: int,
        replacement_length: int,
        patch: Patch,
        context: InsertionContext,
    ) -> Tuple[gtirb.CodeBlock, int]:
        """
        Invokes a patch at a concrete location and applies its results to the
        target module.
        :param func: The function to insert at.
        :param block: The block to insert at.
        :param offset: The offset within the block to insert at.
        :param patch: The patch to invoke.
        :param context: The InsertionContext to pass to the patch.
        :returns: A tuple with: the block that ends the patch and the number
                  of bytes inserted.
        """

        available_registers = list(self._isa.all_registers())
        clobbered_registers = set()
        snippets = []
        stack_adjustment = 0

        for clobber in patch.constraints.clobbers_registers:
            reg = self._isa.get_register(clobber)
            available_registers.remove(reg)
            clobbered_registers.add(reg)

        scratch_registers = available_registers[
            : patch.constraints.scratch_registers
        ]
        clobbered_registers.update(scratch_registers)

        if patch.constraints.preserve_caller_saved_registers:
            clobbered_registers.update(self._isa.caller_saved_registers())

        # TODO: If align_stack was set too, we're going to end up doing
        #       some redundant work.
        if clobbered_registers or patch.constraints.clobbers_flags:
            if self._isa.red_zone_size() and self._leaf_functions[func.uuid]:
                stack_adjustment += self._isa.red_zone_size()
                snippets.append(self._isa.preserve_red_zone())

        if patch.constraints.clobbers_flags:
            stack_adjustment += self._isa.pointer_size()
            snippets.append(self._isa.save_flags())

        for reg in sorted(clobbered_registers, key=lambda reg: reg.name):
            stack_adjustment += self._isa.pointer_size()
            snippets.append(self._isa.save_register(reg))

        if patch.constraints.align_stack:
            # TODO: We don't know how much the stack may be adjusted by the
            #       snippet.
            stack_adjustment = None
            snippets.append(self._isa.align_stack())

        asm = patch.get_asm(
            dataclasses.replace(context, stack_adjustment=stack_adjustment),
            *scratch_registers,
        )
        if not asm:
            return block, 0

        self._patch_id += 1

        assembler = _Assembler(
            self._module, self._patch_id, self._symbols_by_name
        )
        for snippet in snippets:
            assembler.assemble(snippet[0].code, snippet[0].x86_syntax)
        try:
            assembler.assemble(asm, patch.constraints.x86_syntax)
        except mcasm.assembler.AsmSyntaxError as err:
            self._log_patch_error(asm, patch, self._patch_id, err)
            raise
        for snippet in reversed(snippets):
            assembler.assemble(snippet[1].code, snippet[1].x86_syntax)
        assembler.finalize()

        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug("Applying %s at %s+%s", patch, block, offset)
            self._logger.debug("  Before:")
            show_block_asm(block, decoder=self._decoder, logger=self._logger)

        _modify_block_insert(
            block,
            offset,
            replacement_length,
            bytes(assembler.data),
            assembler.blocks,
            assembler.cfg,
            assembler.symbolic_expressions,
            assembler.local_symbols.values(),
            assembler.proxies,
            self._functions_by_block,
        )

        if "functionBlocks" in self._module.aux_data:
            function_blocks = self._module.aux_data["functionBlocks"].data
            if func.uuid in function_blocks:
                function_blocks[func.uuid].update(
                    b for b in assembler.blocks if b.module == self._module
                )

        self._functions_by_block.update(
            {
                b: func.uuid
                for b in assembler.blocks
                if b.module == self._module
            }
        )

        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug("  After:")
            show_block_asm(block, decoder=self._decoder, logger=self._logger)
            for patch_block in assembler.blocks:
                show_block_asm(
                    patch_block, decoder=self._decoder, logger=self._logger
                )

        return (
            assembler.blocks[-1] if assembler.blocks else block,
            len(assembler.data),
        )

    def get_or_insert_extern_symbol(
        self, name: str, libname: str, preload: bool = False
    ) -> gtirb.Symbol:
        """
        Gets a symbol by name, creating it as an extern symbol if it isn't
        already in the module.
        :param name: The name of the symbol.
        :param libname: The name of the shared library the symbol is from.
        :param preload: Insert the library dependency at the beginning of the
                        libraries aux data table, similar to LD_PRELOAD. ELF
                        only, optional.
        """
        name = decorate_extern_symbol(self._module, name)

        sym = next(
            (sym for sym in self._module.symbols if sym.name == name), None
        )
        if sym:
            return sym

        proxy = gtirb.ProxyBlock()
        sym = gtirb.Symbol(name, payload=proxy)
        self._module.symbols.add(sym)

        if self._module.file_format == gtirb.Module.FileFormat.PE:
            # The pprinter uses symbol forwarding to figure out
            # how to pprint imported symbols.  The fact that it has an entry in
            # the symbolForwarding table will trigger this.  Typically I think
            # the local symbol (pre-forwarding) is something generated by
            # ddisasm, with the proper import name being in the forwarding
            # table.  Here, for a new symbol, having it forward to itself is
            # what we actually want.
            self._module.aux_data["symbolForwarding"].data[sym] = sym
            # May not be neccessary, but should be done for IR consistency
            self._module.aux_data["peImportedSymbols"].data.append(sym)
            if "peImportEntries" in self._module.aux_data:
                self._module.aux_data["peImportEntries"].data.append(
                    (0, -1, name, libname)
                )

        if "libraries" in self._module.aux_data:
            if preload:
                self._module.aux_data["libraries"].data.insert(0, libname)
            else:
                self._module.aux_data["libraries"].data.append(libname)

        return sym

    def _apply_insertions(
        self,
        insertions: Sequence[_Insertion],
        func: gtirb_functions.Function,
        block: gtirb.CodeBlock,
    ) -> None:
        """
        Applies all of the patches that apply to a single block.
        """

        instructions = None
        if any(
            insertion.scope._needs_disassembly() for insertion in insertions
        ):
            instructions = tuple(self._decoder.get_instructions(block))

        # Determine the insertion location for each patch.
        # TODO: This is where bubbling will get hooked in, but for now
        #       always insert at the first potential offset.
        insertions_and_offsets = []
        for insertion in insertions:
            offset = next(
                insertion.scope._potential_offsets(func, block, instructions)
            )
            insertions_and_offsets.append((insertion, offset))

        # Now sort all of the insertions by their offsets. Python uses a
        # stable sort so that this will still be deterministic for ties.
        insertions_and_offsets.sort(key=operator.itemgetter(1))

        # Assert that we don't have any replacements and insertions that
        # overlap.
        last_end = 0
        for insertion, offset in insertions_and_offsets:
            assert offset >= last_end, "Insertions and replacements overlap"
            last_end = offset + insertion.scope._replacement_length()

        actual_block = block
        total_insert_len = 0
        for insertion, offset in insertions_and_offsets:
            block_delta = actual_block.offset - block.offset
            actual_block, insert_len = self._invoke_patch(
                func,
                actual_block,
                offset + total_insert_len - block_delta,
                insertion.scope._replacement_length(),
                insertion.patch,
                InsertionContext(self._module, func, block, offset),
            )
            total_insert_len += (
                insert_len - insertion.scope._replacement_length()
            )

    def _insert_function_stub(
        self, sym: gtirb.Symbol, block: gtirb.CodeBlock
    ) -> None:
        """
        Inserts a stub that just contains a return instruction and
        corresponding return edge.
        """
        assert sym.referent == block
        assert block.size == 0

        func_uuid = uuid.uuid4()

        # Our code block needs something in it for now, along with a return
        # edge for callers to update. The actual instruction in the block
        # doesn't matter, so we'll use a nop.
        nop_encoding = self._isa.nop()
        block.size = len(nop_encoding)

        bi = gtirb.ByteInterval(contents=nop_encoding, blocks=[block])
        sect = next(
            sect
            for sect in self._module.sections
            if sect.name == _text_section_name(self._module)
        )
        sect.byte_intervals.add(bi)

        return_proxy = gtirb.ProxyBlock()
        self._module.proxies.add(return_proxy)
        self._module.ir.cfg.add(
            gtirb.Edge(
                source=block,
                target=return_proxy,
                label=gtirb.Edge.Label(type=gtirb.Edge.Type.Return),
            )
        )

        # TODO: Should there be a mechanism for configuring this?
        if "elfSymbolInfo" in self._module.aux_data:
            self._module.aux_data["elfSymbolInfo"].data[sym] = (
                0,
                "FUNC",
                "GLOBAL",
                "DEFAULT",
                0,
            )

        if "functionEntries" in self._module.aux_data:
            self._module.aux_data["functionEntries"].data[func_uuid] = {block}

        if "functionBlocks" in self._module.aux_data:
            self._module.aux_data["functionBlocks"].data[func_uuid] = {block}

        if "functionNames" in self._module.aux_data:
            self._module.aux_data["functionNames"].data[func_uuid] = sym

        self._functions_by_block[block] = func_uuid

    def _apply_function_insertion(
        self, sym: gtirb.Symbol, block: gtirb.CodeBlock, patch: Patch
    ) -> None:
        assert sym.referent == block
        assert (
            not patch.constraints.clobbers_registers
        ), "function patches should not set clobbers_registers"
        assert (
            not patch.constraints.scratch_registers
        ), "function patches should not set scratch_registers"
        assert (
            not patch.constraints.preserve_caller_saved_registers
        ), "function patches should not set preserve_caller_saved_registers"
        assert (
            not patch.constraints.clobbers_flags
        ), "function patches should not set clobbers_flags"
        assert (
            not patch.constraints.align_stack
        ), "function patches should not set align_stack"

        func = gtirb_functions.Function(
            self._functions_by_block[block], {block}, {block}, {sym}, set()
        )
        context = InsertionContext(self._module, func, block, 0)

        self._invoke_patch(func, block, 0, block.size, patch, context)

    def _validate_offset_and_length(
        self, block: gtirb.CodeBlock, offset: int, length: int
    ) -> None:
        """
        Validate that an offset and length fall within a code block and, if
        expensive assertions are enabled, verify that they fall on instruction
        boundaries.
        """
        assert 0 <= offset <= block.size
        assert length >= 0
        assert offset + length <= block.size

        if self._expensive_assertions:
            legal_offsets = {
                0,
                *itertools.accumulate(
                    inst.size for inst in self._decoder.get_instructions(block)
                ),
            }
            assert (
                offset in legal_offsets
            ), f"offset {offset} is not an instruction boundary"
            assert (
                offset + length in legal_offsets
            ), f"offset {offset}+{length} is not an instruction boundary"

    def register_insert(self, scope: Scope, patch: Patch) -> None:
        """
        Registers a patch to be inserted.
        :param scope: Where should the patch be placed?
        :param patch: The patch to be inserted.
        """
        self._insertions.append(_Insertion(scope, patch))

    def register_insert_function(
        self, name: str, patch: Patch
    ) -> gtirb.Symbol:
        """
        Registers a patch to be inserted as a function.
        :param name: The name of the function to be inserted.
        :param patch: The patch to be inserted.
        :returns: The new function symbol.
        """
        block = gtirb.CodeBlock()
        # TODO: This assumes the symbol isn't present already in the module,
        #       which may be a problem if you have an existing external
        #       symbol that you want to provide a definition for.
        sym = gtirb.Symbol(name, payload=block)
        sym.referent = block
        # TODO: Should we be adding the symbol here?
        self._module.symbols.add(sym)
        self._symbols_by_name[sym.name] = sym
        self._function_insertions.append(_FunctionInsertion(sym, block, patch))
        return sym

    def insert_at(
        self,
        function: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        offset: int,
        patch: Patch,
    ) -> None:
        """
        Inserts a patch at a specific location in the binary. This is not
        subject to bubbling.
        """
        self._validate_offset_and_length(block, offset, 0)
        self.register_insert(
            _SpecificLocationScope(function, block, offset), patch
        )

    def replace_at(
        self,
        function: gtirb_functions.Function,
        block: gtirb.CodeBlock,
        offset: int,
        length: int,
        patch: Patch,
    ) -> None:
        """
        Inserts a patch at a specific location in the binary, replacing the
        instructions as specified. This is not subject to bubbling.
        """
        self._validate_offset_and_length(block, offset, length)
        self.register_insert(
            _SpecificLocationScope(function, block, offset, length), patch
        )

    def apply(self) -> None:
        """
        Applies all of the patches to the module.
        """

        with prepare_for_rewriting(self._module, self._isa.nop()):
            self._functions_by_block = {
                block: func.uuid
                for func in self._functions
                for block in func.get_all_blocks()
            }

            for func in self._function_insertions:
                self._insert_function_stub(func.symbol, func.block)

            for func in self._function_insertions:
                self._apply_function_insertion(
                    func.symbol, func.block, func.patch
                )

            for f in self._functions:
                func_insertions = [
                    insertion
                    for insertion in self._insertions
                    if insertion.scope._function_matches(self._module, f)
                ]
                if not func_insertions:
                    continue

                # Iterate over initial function blocks; ignore added blocks
                # from patches.
                for b in tuple(f.get_all_blocks()):
                    block_insertions = [
                        insertion
                        for insertion in func_insertions
                        if insertion.scope._block_matches(self._module, f, b)
                    ]
                    if not block_insertions:
                        continue

                    self._apply_insertions(block_insertions, f, b)

        # Remove CFI directives, since we will most likely be invalidating
        # most (or all) of them.
        # TODO: can we not do this?
        if "cfiDirectives" in self._module.aux_data:
            del self._module.aux_data["cfiDirectives"]



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import itertools
from typing import Dict, List, Optional

import gtirb
import mcasm

from .assembly import X86Syntax
from .utils import _is_elf_pie, _target_triple


class _Assembler:
    """
    Assembles the assembly from a patch and its prologue/epilogue, creating
    control flow graph as it goes.

    :ivar ~.data: The encoded instructions from assembling the patch.
    :ivar ~.cfg: The control flow graph for the patch.
    :ivar ~.blocks: All code blocks, in order, for the patch.
    :ivar ~.symbolic_expressions: A map of offset to symbolic expression,
                                  relative to the data position.
    :ivar ~.local_symbols: Symbols that were created in the patch.
    :ivar ~.proxies: Proxy blocks that represent unknown targets.
    """

    def __init__(
        self,
        module: gtirb.Module,
        patch_id: int,
        module_symbols: Dict[str, gtirb.Symbol],
    ) -> None:
        """
        :param module: The module the patch will be inserted into.
        :param patch_id: A unique integer for this patch to avoid ambiguous
                         symbol names.
        :param module_symbols: A map from symbol name to symbol, to speed up
                               name lookups. This may be updated by the
                               assembler upon cache misses.
        """
        self._module = module
        self._patch_id = patch_id
        self.data = bytearray()
        self.cfg = gtirb.CFG()
        self.blocks = [gtirb.CodeBlock()]
        self.symbolic_expressions = {}
        self.local_symbols = {}
        self._module_symbols = module_symbols
        self.section_name = None
        self.proxies = set()

    @property
    def entry_block(self) -> gtirb.CodeBlock:
        return self.blocks[0]

    @property
    def current_block(self) -> gtirb.CodeBlock:
        return self.blocks[-1]

    def assemble(self, asm: str, x86_syntax: X86Syntax) -> None:
        """
        Assembles additional assembly into this chunk, continuing where the
        last call to assemble left off.
        """
        assembler = mcasm.Assembler(_target_triple(self._module))
        assembler.x86_syntax = x86_syntax

        events = assembler.assemble(asm)

        for event in events:
            if event["kind"] == "label":
                self._precreate_defined_label(event["symbol"])

        for event in events:
            if event["kind"] == "instruction":
                self._assemble_instruction(
                    bytes.fromhex(event["data"]),
                    event["inst"],
                    event["fixups"],
                )
            elif event["kind"] == "label":
                self._assemble_label(event["symbol"])
            elif event["kind"] == "changeSection":
                self._assemble_change_section(event["section"])
            else:
                assert False, f"Unsupported assembler event: {event['kind']}"

    def _precreate_defined_label(self, symbol: dict) -> None:
        label_name = symbol["name"]

        # If the symbol is temporary in LLVM's eyes, we will append our patch
        # id to it in order to make the symbol name unique.
        symbol_name = label_name
        if symbol["isTemporary"]:
            symbol_name += f"_{self._patch_id}"

        assert (
            label_name not in self.local_symbols
            and symbol_name not in self._module_symbols
        ), f"{symbol_name} defined multiple times"

        label_sym = gtirb.Symbol(name=symbol_name, payload=gtirb.CodeBlock())
        self.local_symbols[label_name] = label_sym

    def _assemble_label(self, symbol: dict) -> None:
        label_sym = self.local_symbols[symbol["name"]]
        label_block = label_sym.referent
        assert label_block

        label_block.offset = (
            self.current_block.offset + self.current_block.size
        )
        self.cfg.add(
            gtirb.Edge(
                source=self.current_block,
                target=label_block,
                label=gtirb.Edge.Label(type=gtirb.Edge.Type.Fallthrough),
            )
        )

        self.blocks.append(label_block)

    def _assemble_change_section(self, section: dict) -> None:
        # We don't have any work to do here, but we want to make sure that
        # patches don't try to insert anywhere except the text section.
        assert section["kind"][
            "isText"
        ], "Sections other than .text are not supported"
        self.section_name = section["name"]

    def _assemble_instruction(
        self, data: bytes, inst: dict, fixups: List[dict]
    ) -> None:
        for fixup in fixups:
            pos = len(self.data) + fixup["offset"]
            self.symbolic_expressions[pos] = self._fixup_to_symbolic_operand(
                fixup, data
            )

        self.data += data
        self.current_block.size += len(data)

        if inst["desc"]["isReturn"]:
            proxy = gtirb.ProxyBlock()
            self.proxies.add(proxy)
            self.cfg.add(
                gtirb.Edge(
                    source=self.current_block,
                    target=proxy,
                    label=gtirb.Edge.Label(type=gtirb.Edge.Type.Return),
                )
            )

            next_block = gtirb.CodeBlock(
                offset=self.current_block.offset + self.current_block.size
            )
            self.blocks.append(next_block)

        elif inst["desc"]["isCall"] or inst["desc"]["isBranch"]:
            assert not inst["desc"][
                "isIndirectBranch"
            ], "Indirect branches are not yet supported"

            assert len(fixups) == 1
            target_fixup = self._fixup_to_symbolic_operand(fixups[0], data)
            assert isinstance(target_fixup, gtirb.SymAddrConst)
            assert target_fixup.offset == 0

            next_block = gtirb.CodeBlock(
                offset=self.current_block.offset + self.current_block.size
            )

            if isinstance(target_fixup.symbol.referent, gtirb.CfgNode):
                if inst["desc"]["isCall"]:
                    edge_label = gtirb.Edge.Label(type=gtirb.Edge.Type.Call)
                elif inst["desc"]["isBranch"]:
                    edge_label = gtirb.Edge.Label(
                        type=gtirb.Edge.Type.Branch,
                        conditional=inst["desc"]["isConditionalBranch"],
                    )

                self.cfg.add(
                    gtirb.Edge(
                        source=self.current_block,
                        target=target_fixup.symbol.referent,
                        label=edge_label,
                    )
                )

            # Currently we assume that all calls can return and that they need
            # a fallthrough edge.
            if inst["desc"]["isCall"] or inst["desc"]["isConditionalBranch"]:
                self.cfg.add(
                    gtirb.Edge(
                        source=self.current_block,
                        target=next_block,
                        label=gtirb.Edge.Label(
                            type=gtirb.Edge.Type.Fallthrough
                        ),
                    )
                )

            self.blocks.append(next_block)

    def _fixup_to_symbolic_operand(
        self, fixup: dict, encoding: bytes
    ) -> gtirb.SymAddrConst:
        """
        Converts an LLVM fixup to a GTIRB SymbolicExpression.
        """
        expr = fixup["value"]

        # LLVM will automatically add a negative value to make the expression
        # be PC-relative. We don't care about that and just want to unwrap it.
        if (
            "IsPCRel" in fixup["flags"]
            and expr["kind"] == "binaryExpr"
            and expr["opcode"] == "Add"
            and expr["rhs"]["kind"] == "constant"
            and fixup["offset"] - expr["rhs"]["value"] == len(encoding)
        ):
            expr = expr["lhs"]

        # TODO: Do we need to support more fixup types? GTIRB only supports
        #       two forms:
        #       - Sym + Offset
        #       - (Sym1 - Sym2) / Scale + Offset
        assert (
            expr["kind"] == "symbolRef"
        ), "Only simple simple references are currently supported"

        name = expr["symbol"]["name"]
        sym = self._symbol_lookup(name)
        assert sym, f"{name} is an undefined symbol reference"

        attributes = set()
        if "variantKind" in expr:
            if expr["variantKind"] == "PLT":
                attributes.add(gtirb.SymbolicExpression.Attribute.PltRef)
            else:
                assert False, f"Unsupported variantKind: {expr['variantKind']}"
        elif _is_elf_pie(self._module) and isinstance(
            sym.referent, gtirb.ProxyBlock
        ):
            attributes.add(gtirb.SymbolicExpression.Attribute.PltRef)

        return gtirb.SymAddrConst(0, sym, attributes)

    def _symbol_lookup(self, name: str) -> Optional[gtirb.Symbol]:
        """
        Looks up a symbol by name.

        :param name: The symbol's name.
        """

        sym = self.local_symbols.get(name, None)
        if sym:
            return sym

        sym = self._module_symbols.get(name, None)
        if sym and sym.module == self._module:
            return sym

        sym = next(
            (sym for sym in self._module.symbols if sym.name == name), None
        )
        if sym:
            self._module_symbols[sym.name] = sym
            return sym

        return None

    def _remove_empty_blocks(self) -> None:
        final_blocks = []
        for _, group in itertools.groupby(self.blocks, key=lambda b: b.offset):
            *extra_blocks, main_block = group
            assert main_block.size or main_block == self.blocks[-1]

            for extra_block in extra_blocks:
                assert not extra_block.size

                for edge in list(self.cfg.in_edges(extra_block)):
                    self.cfg.discard(edge)
                    if edge.source not in extra_blocks:
                        assert edge.source != main_block
                        self.cfg.add(edge._replace(target=main_block))

                # Our extra block should only have a single fallthrough edge
                # that is to another extra block or the main block.
                for edge in list(self.cfg.out_edges(extra_block)):
                    assert edge.label.type == gtirb.Edge.Type.Fallthrough
                    assert (
                        edge.target in extra_blocks
                        or edge.target == main_block
                    )
                    self.cfg.discard(edge)

                for sym in self.local_symbols.values():
                    if sym.referent == extra_block:
                        sym.referent = main_block

            final_blocks.append(main_block)

        self.blocks = final_blocks

    def finalize(self) -> None:
        """
        Finalizes the assembly contents and validates that there are no
        undefined symbols referenced.
        """

        self._remove_empty_blocks()



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import dataclasses
from typing import Callable, Iterable, List, Optional, Union

import gtirb

from ..assembly import Constraints, InsertionContext, Patch, X86Syntax
from ..isa import CallingConventionDesc, _get_isa


class CallPatch(Patch):
    """
    Inserts a call to a function with literal arguments.
    """

    ActualArgumentValue = Union[int, gtirb.Symbol]
    ArgumentValue = Union[
        Callable[[InsertionContext], "ActualArgumentValue"],
        "ActualArgumentValue",
    ]

    @dataclasses.dataclass
    class _PassedArg:
        value: "CallPatch.ArgumentValue"
        reg: Optional[str]

    def __init__(
        self,
        sym: gtirb.Symbol,
        args: Iterable["ArgumentValue"] = (),
        conv: Optional[CallingConventionDesc] = None,
        **constraint_kwargs,
    ):
        """
        Initializes a call patch.

        sym: The symbol to call.
        args: An iterable of arguments to be passed to the function. They must
              be either integers or Symbols.
        conv: The calling convention description to use for the call. If
              absent, the default ABI calling convention is used.
        constraint_kwargs: Additional keyword arguments to be passed to the
                         patch's constraints.
        """

        assert sym.module, "Symbol must be in a module"
        assert sym.module.isa in (gtirb.Module.ISA.IA32, gtirb.Module.ISA.X64)

        self._isa = _get_isa(sym.module)
        if conv:
            self._cconv = conv
        else:
            self._cconv = self._isa.calling_convention()

        self.sym = sym
        self._args = self._create_passed_args(self._cconv, args)

        super().__init__(
            Constraints(
                x86_syntax=X86Syntax.INTEL,
                clobbers_flags=True,
                clobbers_registers={arg.reg for arg in self._args if arg.reg},
                **constraint_kwargs,
            )
        )

    def _create_passed_args(
        self, conv: CallingConventionDesc, args: Iterable["ArgumentValue"]
    ) -> List["_PassedArg"]:
        """
        Assigns arguments to registers (or the stack) and creates the string
        value of the argument.
        """
        remaining_regs = list(conv.registers)

        passed_args = []
        for arg in args:
            assert (
                callable(arg)
                or isinstance(arg, gtirb.Symbol)
                or isinstance(arg, int)
            )

            passed_args.append(
                self._PassedArg(
                    arg, remaining_regs.pop(0) if remaining_regs else None
                )
            )

        return passed_args

    def get_asm(self, insertion_context: InsertionContext) -> str:
        lines = []

        stack_slot_size = self._isa.pointer_size()
        stack_reg = self._isa.stack_register()

        stack_size = sum(stack_slot_size for arg in self._args if not arg.reg)
        stack_padding = stack_size % self._cconv.stack_alignment
        if stack_padding:
            lines.append(f"sub {stack_reg}, {stack_padding}")

        for arg in reversed(self._args):
            arg_value = (
                arg.value(insertion_context)
                if callable(arg.value)
                else arg.value
            )

            if isinstance(arg_value, gtirb.Symbol):
                arg_str = arg_value.name
            elif isinstance(arg_value, int):
                arg_str = str(arg_value)

            if arg.reg:
                lines.append(f"mov {arg.reg}, {arg_str}")
            else:
                lines.append(f"push {arg_str}")

        if self._cconv.shadow_space:
            lines.append(f"sub {stack_reg}, {self._cconv.shadow_space}")

        lines.append(f"call {self.sym.name}")

        cleanup_size = self._cconv.shadow_space + stack_padding
        if self._cconv.caller_cleanup:
            cleanup_size += stack_size

        if cleanup_size:
            lines.append(f"add {stack_reg}, {cleanup_size}")

        return "\n".join(lines)



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import dataclasses
from typing import List, Set, Tuple

import gtirb

from .assembly import Register, _AsmSnippet


@dataclasses.dataclass
class CallingConventionDesc:
    """
    Describes an ABI's calling convention.
    """

    registers: Tuple[str, ...]
    """
    Registers used for passing integer or pointer-like values, in the
    order they are to be used. Any remaining arguments will be passed on
    the stack.
    """

    stack_alignment: int
    """
    The required stack alignment for calling a function.
    """

    caller_cleanup: bool
    """
    Is the caller required to do cleanup for arguments pushed onto the stack?
    """

    shadow_space: int = 0
    """
    The amount of space required to be allocated on the stack before calling
    the function. This space is essentially empty can be used by the called
    function.
    """


class _ISA:
    def __init__(self) -> None:
        self._register_map = {}
        for reg in self.all_registers():
            for name in reg.sizes.values():
                self._register_map[name.lower()] = reg

    def save_register(
        self, register: Register
    ) -> Tuple[_AsmSnippet, _AsmSnippet]:
        """
        Generate code required to save a specific register.
        """
        raise NotImplementedError

    def align_stack(self) -> Tuple[_AsmSnippet, _AsmSnippet]:
        """
        Generate code to align the stack to the ABI requirements for a call.
        """
        raise NotImplementedError

    def get_register(self, name: str) -> Register:
        """
        Gets a Register object by its name (or the name of a subregister).
        """
        return self._register_map[name.lower()]

    def save_flags(self) -> Tuple[_AsmSnippet, _AsmSnippet]:
        """
        Generate code required to save the flags register.
        """
        raise NotImplementedError

    def all_registers(self) -> List[Register]:
        """
        Returns all general-purpose registers for the ISA.
        """
        raise NotImplementedError

    def nop(self) -> bytes:
        """
        Returns the encoding of a no-op instruction.
        """
        raise NotImplementedError

    def caller_saved_registers(self) -> Set[Register]:
        """
        Returns the registers that need to be saved by the caller if it wants
        the values preserved across the call.
        """
        raise NotImplementedError

    def pointer_size(self) -> int:
        """
        Returns the size of a pointer on the ISA (which is assumed to match
        the size of general purpose registers).
        """
        raise NotImplementedError

    def red_zone_size(self) -> int:
        """
        Returns the number of bytes that leaf functions are allowed to use on
        the stack (without adjusting the stack pointer).
        """
        return 0

    def preserve_red_zone(self) -> Tuple[_AsmSnippet, _AsmSnippet]:
        """
        Generate code required to preserve the contents of the red zone.
        """
        raise NotImplementedError

    def calling_convention(self) -> CallingConventionDesc:
        """
        Returns a description of the ABI's default calling convention.
        """
        raise NotImplementedError

    def stack_register(self) -> Register:
        """
        Returns the stack pointer register.
        """
        raise NotImplementedError


class _X86_64(_ISA):
    def save_register(
        self, register: Register
    ) -> Tuple[_AsmSnippet, _AsmSnippet]:
        return (
            _AsmSnippet(f"pushq %{register}"),
            _AsmSnippet(f"popq %{register}"),
        )

    def save_flags(self) -> Tuple[_AsmSnippet, _AsmSnippet]:
        # TODO: Replace this with something more efficient.
        return _AsmSnippet("pushfq"), _AsmSnippet("popfq")

    def align_stack(self) -> Tuple[_AsmSnippet, _AsmSnippet]:
        return (
            _AsmSnippet(
                """
                pushq   %rax
                movq    %rsp, %rax
                leaq    -0x80(%rsp), %rsp
                andq    $-0x10, %rsp
                pushq   %rax
                pushq   %rax
            """
            ),
            _AsmSnippet(
                """
                popq    %rax
                movq    %rax, %rsp
                popq    %rax
            """
            ),
        )

    def all_registers(self) -> List[Register]:
        return [
            Register(
                {"8l": "al", "8h": "ah", "16": "ax", "32": "eax", "64": "rax"},
                "64",
            ),
            Register(
                {"8l": "bl", "8h": "bh", "16": "bx", "32": "ebx", "64": "rbx"},
                "64",
            ),
            Register(
                {"8l": "cl", "8h": "ch", "16": "cx", "32": "ecx", "64": "rcx"},
                "64",
            ),
            Register(
                {"8l": "dl", "8h": "dh", "16": "dx", "32": "edx", "64": "rdx"},
                "64",
            ),
            Register(
                {"8l": "sil", "16": "si", "32": "esi", "64": "rsi"}, "64"
            ),
            Register(
                {"8l": "dil", "16": "di", "32": "edi", "64": "rdi"}, "64"
            ),
            Register(
                {"8l": "r8b", "16": "r8w", "32": "r8d", "64": "r8"}, "64"
            ),
            Register(
                {"8l": "r9b", "16": "r9w", "32": "r9d", "64": "r9"}, "64"
            ),
            Register(
                {"8l": "r10b", "16": "r10w", "32": "r10d", "64": "r10"}, "64"
            ),
            Register(
                {"8l": "r11b", "16": "r11w", "32": "r11d", "64": "r11"}, "64"
            ),
            Register(
                {"8l": "r12b", "16": "r12w", "32": "r12d", "64": "r12"}, "64"
            ),
            Register(
                {"8l": "r13b", "16": "r13w", "32": "r13d", "64": "r13"}, "64"
            ),
            Register(
                {"8l": "r14b", "16": "r14w", "32": "r14d", "64": "r14"}, "64"
            ),
            Register(
                {"8l": "r15b", "16": "r15w", "32": "r15d", "64": "r15"}, "64"
            ),
        ]

    def nop(self) -> bytes:
        return b"\x90"

    def pointer_size(self) -> int:
        return 8

    def stack_register(self) -> Register:
        return Register({"16": "sp", "32": "esp", "64": "rsp"}, "64",)


class _X86_64_PE(_X86_64):
    def caller_saved_registers(self) -> Set[Register]:
        return {
            self.get_register(name)
            for name in ("RAX", "RCX", "RDX", "R8", "R9", "R10", "R11")
        }

    def calling_convention(self) -> CallingConventionDesc:
        return CallingConventionDesc(
            registers=("RCX", "RDX", "R8", "R9"),
            stack_alignment=16,
            caller_cleanup=True,
            shadow_space=32,
        )


class _X86_64_ELF(_X86_64):
    def caller_saved_registers(self) -> Set[Register]:
        return {
            self.get_register(name)
            for name in (
                "RAX",
                "RCX",
                "RDX",
                "RSI",
                "RDI",
                "R8",
                "R9",
                "R10",
                "R11",
            )
        }

    def red_zone_size(self) -> int:
        return 128

    def preserve_red_zone(self) -> Tuple[_AsmSnippet, _AsmSnippet]:
        return (
            _AsmSnippet("leaq -128(%rsp), %rsp"),
            _AsmSnippet("leaq +128(%rsp), %rsp"),
        )

    def calling_convention(self) -> CallingConventionDesc:
        return CallingConventionDesc(
            registers=("RDI", "RSI", "RDX", "RCX", "R8", "R9"),
            stack_alignment=16,
            caller_cleanup=True,
        )


def _get_isa(module: gtirb.Module) -> _ISA:
    if module.isa == gtirb.Module.ISA.X64:
        if module.file_format == gtirb.Module.FileFormat.ELF:
            return _X86_64_ELF()
        elif module.file_format == gtirb.Module.FileFormat.PE:
            return _X86_64_PE()

    assert False, f"Unsupported ISA/format: {module.isa}/{module.file_format}"



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import dataclasses
from functools import partial
from typing import Callable, Dict, Optional, Set

import gtirb
import gtirb_functions
import mcasm

from .utils import decorate_extern_symbol


class Register:
    """
    An architecture-specific register.

    Registers are not created directly and are instead given to patches by
    requesting scratch registers in their constraints.
    """

    def __init__(self, sizes: Dict[str, str], default_size: str):
        self.sizes = sizes
        self.default_size = default_size

    def __contains__(self, value) -> bool:
        return value in self.sizes.values()

    def __eq__(self, other) -> bool:
        return (
            self.sizes == other.sizes
            and self.default_size == other.default_size
        )

    def __hash__(self) -> int:
        return hash(self.default_size) ^ hash(
            tuple(sorted(self.sizes.items()))
        )

    def __format__(self, spec: str) -> str:
        """
        Formats the register (or subregister) as its name.

        :param spec: The format type specifier. This can control which
                     subregister name is used.
                     x86-64 supports the following sizes:
                     - 8l, 8h, 16, 32, 64
        """
        if spec:
            return self.sizes[spec]
        return self.sizes[self.default_size]

    @property
    def name(self) -> str:
        return self.sizes[self.default_size]


@dataclasses.dataclass
class InsertionContext:
    """
    A concrete location to insert code at, plus helper utilities.
    """

    module: gtirb.Module
    function: gtirb_functions.Function
    block: gtirb.CodeBlock
    offset: int
    """
    The byte offset of the insertion, relative to the start of the block.
    """

    stack_adjustment: Optional[int] = None
    """
    The amount that the stack has been adjusted to allow for the insertion.
    A value of None means that it is unknown how much it has been affected
    (which can happen with the align_stack constraint).
    """

    def decorate_extern_symbol(self, name: str) -> str:
        return decorate_extern_symbol(self.module, name)


X86Syntax = mcasm.X86Syntax


@dataclasses.dataclass
class Constraints:
    """
    Constraints related to the assembly code in a patch. These can be seen as
    metadata about the actual assembly in a patch and can impact where the
    patch gets placed if bubbling is allowed.
    """

    x86_syntax: X86Syntax = X86Syntax.ATT
    """
    The syntax mode to use for x86 code. This is unused for other ISAs.
    """

    clobbers_flags: bool = False
    """
    Does the assembly clobber the flags register?
    """

    clobbers_registers: Set[str] = dataclasses.field(default_factory=set)
    """
    The general purpose registers that the assembly clobbers.
    """

    scratch_registers: int = 0
    """
    The number of scratch registers that the patch needs. When emitting the
    patch's code, the rewriting context will try to pick free registers. If no
    registers are free at the insertion point, it will generate code to spill
    the registers to the stack before/after the patch.

    The scratch registers will be passed as positional arguments to the patch's
    get_asm method.
    """

    align_stack: bool = False
    """
    Generate code to align the stack to the ABI-defined alignment before
    emitting the patch's code and restore the previous value after. This is
    useful for inserting a call into a module's entrypoint.
    """

    preserve_caller_saved_registers: bool = False
    """
    Spill all caller saved registers as per the ABI based on ISA and ouput
    format (e.g. PE vs ELF).
    """


def patch_constraints(*args, **kwargs):
    """
    Associates a Constraints object with a function that is meant to be used
    as a Patch (see Patch.from_function). The arguments to the decorator are
    used when constructing the Constraints object.
    """

    def constraints_decorator(func):
        func.constraints = Constraints(*args, **kwargs)
        return func

    return constraints_decorator


@dataclasses.dataclass
class _AsmSnippet:
    code: str
    x86_syntax: X86Syntax = X86Syntax.ATT


def _unwrap_callable(func):
    while True:
        yield func

        if isinstance(func, partial):
            func = func.func
        elif hasattr(func, "__wrapper__"):
            func = func.__wrapper__
        else:
            break


def _find_constraints(func):
    for layer in _unwrap_callable(func):
        constraints = getattr(layer, "constraints", None)
        if constraints:
            return constraints
    return None


class Patch:
    """
    A chunk of assembly code to be inserted into a program, along with its
    constraints.
    """

    def __init__(self, constraints: Constraints):
        self.constraints = constraints

    def get_asm(
        self, insertion_context: InsertionContext, *args: Register
    ) -> Optional[str]:
        """
        Returns the assembly string for the patch.

        If the assembly string references symbols, the GTIRB module's symbol
        table will be updated as needed and symbolic expressions will be
        created.

        If None is returned, no insertion takes place.

        :param insertion_context: The concrete location where the code will be
                                  inserted.
        :param args: Any scratch registers requested by the patch.
        """
        raise NotImplementedError

    @classmethod
    def from_function(cls, func: Callable, constraints: Constraints = None):
        """
        Creates a Patch from a callable that has been decorated with the
        @patch_constraints decorator.
        """

        class FuncPatch(Patch):
            def __str__(self) -> str:
                return str(func)

            def get_asm(self, insertion_context, *args):
                return func(insertion_context, *args)

        if not constraints:
            constraints = _find_constraints(func)
            assert (
                constraints
            ), "constraints must be specified when not using the decorator"

        return FuncPatch(constraints)



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import logging
import uuid
from typing import (
    Container,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Set,
    TypeVar,
    Union,
    overload,
)

import capstone
import gtirb
from gtirb_capstone.instructions import GtirbInstructionDecoder

T = TypeVar("T")


class OffsetMapping(MutableMapping[gtirb.Offset, T]):
    """Mapping that allows looking up groups of items by their offset element.

    The keys in this mapping are required to be Offsets. If a non-Offset is
    used as a key, it is assumed to be the element_id of an Offset. In that
    case, the corresponding element is a MutableMapping[int, T] of
    displacements to values for every Offset that has the given element_id. For
    example,
        m = OffsetMapping[str]()
        m[Offset(x, 0)] = "a"     # insert an offset into the map
        m[x] = {1: "b", 2: "c"}   # insert two offsets into the map
        m[x][0] = "d"             # change the value for Offset(x, 0)
        print(m[Offset(x, 1)])    # get the value for Offset(x, 1)
        del m[Offset(x, 2)]       # delete Offset(x, 2) from the map
    """

    def __init__(self, *args, **kw):
        """Create a new OffsetMapping from an iterable and/or keywords."""
        self._data = {}
        self.update(*args, **kw)

    def __len__(self) -> int:
        """Get the number of Offsets stored in this mapping."""
        return sum(len(subdata) for subdata in self._data.values())

    def __iter__(self) -> Iterator[gtirb.Offset]:
        """"Yield the Offsets in this mapping."""
        for elem, subdata in self._data.items():
            for disp in subdata:
                yield gtirb.Offset(elem, disp)

    @overload
    def __getitem__(self, key: gtirb.Offset) -> T:
        ...

    @overload
    def __getitem__(self, key: gtirb.Node) -> MutableMapping[int, T]:
        ...

    def __getitem__(self, key):
        """Get the value for an Offset or dictionary for an element_id."""
        if isinstance(key, gtirb.Offset):
            elem, disp = key
            if elem in self._data and disp in self._data[elem]:
                return self._data[elem][disp]
        return self._data[key]

    @overload
    def __setitem__(self, key: gtirb.Offset, value: T) -> None:
        ...

    @overload
    def __setitem__(self, key: gtirb.Node, value: Mapping[int, T]) -> None:
        ...

    def __setitem__(self, key, value):
        """Set the value for an Offset, or several Offsets given an element."""
        if isinstance(key, gtirb.Offset):
            elem, disp = key
            if elem not in self._data:
                self._data[elem] = {}
            self._data[elem][disp] = value
        elif not isinstance(value, Mapping):
            raise ValueError("not a mapping: %r" % value)
        else:
            self._data.setdefault(key, {}).update(value)

    def __delitem__(self, key: Union[gtirb.Offset, gtirb.Node]) -> None:
        """Delete the mapping for an Offset or all Offsets given an element."""
        if isinstance(key, gtirb.Offset):
            elem, disp = key
            if elem not in self._data or disp not in self._data[elem]:
                raise KeyError(key)
            del self._data[elem][disp]
        else:
            del self._data[key]


def _target_triple(module: gtirb.Module) -> str:
    """
    Generate the appropriate LLVM target triple for a GTIRB Module.
    """

    if module.isa == gtirb.Module.ISA.X64:
        arch = "x86_64"
    elif module.isa == gtirb.Module.ISA.IA32:
        arch = "i386"
    elif module.isa == gtirb.Module.ISA.ARM:
        arch = "arm"
    elif module.isa == gtirb.Module.ISA.ARM64:
        arch = "arm64"
    else:
        assert False, f"Unsupported ISA: {module.isa}"

    if module.file_format == gtirb.Module.FileFormat.ELF:
        vendor = "pc"
        os = "linux"
    elif module.file_format == gtirb.Module.FileFormat.PE:
        vendor = "pc"
        os = "win32"
    else:
        assert False, f"Unsupported file format: {module.file_format}"

    return f"{arch}-{vendor}-{os}"


def _nonterminator_instructions(
    block: gtirb.CodeBlock, disassembly: Sequence[capstone.CsInsn]
) -> Iterator[capstone.CsInsn]:
    """
    Yields all instructions in a block of diassembly except for the terminator,
    if present.
    """
    if all(
        edge.label.type == gtirb.Edge.Type.Fallthrough
        for edge in block.outgoing_edges
    ):
        yield from disassembly
    else:
        yield from disassembly[:-1]


def show_block_asm(
    block: gtirb.CodeBlock,
    arch: gtirb.Module.ISA = None,
    logger=logging.getLogger(),
    decoder=None,
) -> None:
    """
    Disassemble and print the contents of a code block using the given
    architecture. If no architecture is given, it is taken from the block's
    module. If the block is not in a module, the function throws an error.
    """

    if decoder is None:
        if arch is None:
            if block.module is None:
                raise ValueError("Undefined architecture")
            arch = block.byte_interval.section.module.isa
        decoder = GtirbInstructionDecoder(arch)

    if block.contents:
        for i in decoder.get_instructions(block):
            logger.debug("\t0x%x:\t%s\t%s", i.address, i.mnemonic, i.op_str)


def _substitute_block(
    old_block: gtirb.CodeBlock,
    new_block: gtirb.CodeBlock,
    cfg: gtirb.CFG,
    symbols: Iterable[gtirb.Symbol],
) -> None:
    """
    Substitutes one block for another by adjusting the CFG and any symbols
    pointing at the block.
    """

    for edge in set(cfg.in_edges(old_block)):
        cfg.discard(edge)
        cfg.add(edge._replace(target=new_block))

    for edge in set(cfg.out_edges(old_block)):
        cfg.discard(edge)
        cfg.add(edge._replace(source=new_block))

    for sym in symbols:
        if sym.referent == old_block:
            sym.referent = new_block


def _is_fallthrough_edge(edge: gtirb.Edge) -> bool:
    """Determines if an edge is a fall-through edge."""
    return edge.label and edge.label.type == gtirb.Edge.Type.Fallthrough


def _is_return_edge(edge: gtirb.Edge) -> bool:
    return edge.label and edge.label.type == gtirb.Edge.Type.Return


def _is_call_edge(edge: gtirb.Edge) -> bool:
    return edge.label and edge.label.type == gtirb.Edge.Type.Call


def _block_return_edges(block: gtirb.CodeBlock) -> Set[gtirb.Edge]:
    return {edge for edge in block.outgoing_edges if _is_return_edge(edge)}


def _block_fallthrough_targets(block: gtirb.CodeBlock) -> Set[gtirb.CodeBlock]:
    return {
        edge.target
        for edge in block.outgoing_edges
        if _is_fallthrough_edge(edge)
    }


def _get_function_blocks(
    module: gtirb.Module, func_uuid: uuid.UUID
) -> Set[gtirb.CodeBlock]:
    """
    Gets all blocks associated with a function.
    """
    if "functionBlocks" in module.aux_data:
        return module.aux_data["functionBlocks"].data[func_uuid]
    else:
        return set()


def _add_return_edges_to_one_function(
    module: gtirb.Module,
    func_uuid: uuid.UUID,
    return_target: gtirb.CodeBlock,
    new_cfg: gtirb.CFG,
) -> None:
    """
    Adds a new return edge to all returns in the function.
    """
    for block in _get_function_blocks(module, func_uuid):
        return_edges = _block_return_edges(block)
        if not return_edges:
            continue

        for return_edge in return_edges:
            if isinstance(return_edge.target, gtirb.ProxyBlock):
                # We are intentionally leaving the proxy block in the module's
                # proxies.
                block.ir.cfg.discard(return_edge)

        new_cfg.add(
            gtirb.Edge(
                source=block,
                target=return_target,
                label=gtirb.Edge.Label(type=gtirb.Edge.Type.Return),
            )
        )


def _add_return_edges_for_patch_calls(
    module: gtirb.Module,
    functions_by_block: Dict[gtirb.CodeBlock, uuid.UUID],
    new_cfg: gtirb.CFG,
) -> None:
    """
    Finds all of the call edges added by the patch and adds new return edges
    to the callee.
    """
    call_edges = {edge for edge in new_cfg if _is_call_edge(edge)}
    for call_edge in call_edges:
        func_uuid = functions_by_block.get(call_edge.target, None)
        if not func_uuid:
            continue

        fallthrough_targets = [
            edge.target
            for edge in new_cfg
            if edge.source == call_edge.source and _is_fallthrough_edge(edge)
        ]
        if not fallthrough_targets:
            continue

        # Because the assembler is generating this input, we can assert that
        # there's only the single fallthrough target.
        assert len(fallthrough_targets) == 1
        _add_return_edges_to_one_function(
            module, func_uuid, fallthrough_targets[0], new_cfg
        )


def _update_patch_return_edges_to_match(
    block: gtirb.CodeBlock,
    functions_by_block: Dict[gtirb.CodeBlock, uuid.UUID],
    new_cfg: gtirb.CFG,
    new_proxy_blocks: Set[gtirb.ProxyBlock],
) -> None:
    """
    Finds all return edges in a patch and updates them to match the function
    being inserted into.
    """
    func_uuid = functions_by_block.get(block, None)
    if not func_uuid:
        return

    return_targets = set()
    for func_block in _get_function_blocks(block.module, func_uuid):
        return_targets.update(
            edge.target
            for edge in func_block.outgoing_edges
            if _is_return_edge(edge)
            and not isinstance(edge.target, gtirb.ProxyBlock)
        )

    if not return_targets:
        return

    patch_return_edges = {edge for edge in new_cfg if _is_return_edge(edge)}
    for edge in patch_return_edges:
        assert isinstance(edge.target, gtirb.ProxyBlock)
        new_cfg.discard(edge)
        new_proxy_blocks.discard(edge.target)
        for target in return_targets:
            new_cfg.add(
                gtirb.Edge(
                    source=edge.source,
                    target=target,
                    label=gtirb.Edge.Label(type=gtirb.Edge.Type.Return),
                )
            )


def _update_return_edges_from_removing_call(
    call_edge: gtirb.Edge,
    fallthrough_targets: Container[gtirb.CodeBlock],
    functions_by_block: Dict[gtirb.CodeBlock, uuid.UUID],
    new_cfg: gtirb.CFG,
) -> None:
    """
    Updates return edges due to removing a call edge.
    """
    if isinstance(call_edge.target, gtirb.ProxyBlock):
        return

    func_uuid = functions_by_block.get(call_edge.target, None)
    if not func_uuid:
        return

    for block in _get_function_blocks(call_edge.target.module, func_uuid):
        return_edges = _block_return_edges(block)
        if not return_edges:
            continue

        remaining_edges = set()
        for edge in return_edges:
            if edge.target in fallthrough_targets:
                block.ir.cfg.discard(edge)
            else:
                remaining_edges.add(edge)

        if not remaining_edges:
            proxy = gtirb.ProxyBlock()
            new_cfg.add(
                gtirb.Edge(
                    source=block,
                    target=proxy,
                    label=gtirb.Edge.Label(type=gtirb.Edge.Type.Return),
                )
            )
            block.module.proxies.add(proxy)


def _update_return_edges_from_changing_fallthrough(
    call_edge: gtirb.Edge,
    fallthrough_targets: Container[gtirb.CodeBlock],
    functions_by_block: Dict[gtirb.CodeBlock, uuid.UUID],
    new_fallthrough: gtirb.CodeBlock,
    new_cfg: gtirb.CFG,
) -> None:
    """
    Updates all return edges in a function that point to a given return target
    to point to a new target.
    """
    if isinstance(call_edge.target, gtirb.ProxyBlock):
        return

    target_func_uuid = functions_by_block.get(call_edge.target, None)
    if not target_func_uuid:
        return

    for target_block in _get_function_blocks(
        call_edge.target.module, target_func_uuid
    ):
        for edge in _block_return_edges(target_block):
            if edge.target in fallthrough_targets:
                target_block.ir.cfg.discard(edge)
                new_cfg.add(edge._replace(target=new_fallthrough))


def _modify_block_insert(
    block: gtirb.CodeBlock,
    offset: int,
    replacement_length: int,
    new_bytes: bytes,
    new_blocks: List[gtirb.CodeBlock],
    new_cfg: gtirb.CFG,
    new_symbolic_expressions: Dict[int, gtirb.SymbolicExpression],
    new_symbols: Iterable[gtirb.Symbol],
    new_proxy_blocks: Set[gtirb.ProxyBlock],
    functions_by_block: Dict[gtirb.CodeBlock, uuid.UUID],
) -> None:
    """
    Insert bytes into a block and adjusts the IR as needed.
    :param block: The code block to insert into.
    :param offset: The byte offset into the code block.
    :param replacement_length: The number of bytes after `offset` that should
                               be removed.
    :param new_bytes: The new content to be inserted.
    :param new_blocks: The blocks corresponding to the new content.
    :param new_cfg: The CFG corresponding to the new content.
    :param new_symbolic_expressions: The symbolic expressions corresponding to
                                     the new content, with the keys relative
                                     to the start of `new_bytes`.
    :param new_symbols: Any symbols that may be defined in the new content.
    :param new_proxy_blocks: All proxy blocks used by the new content.
    :param functions_by_block: Map from code block to containing function UUID.
    """

    assert block.size
    assert 0 <= offset <= block.size
    assert 0 <= offset + replacement_length <= block.size
    assert replacement_length >= 0
    assert new_bytes
    assert new_blocks
    assert block not in new_blocks
    assert new_blocks[0].size, "must have at least one non-empty block"
    assert all(
        new_block.size for new_block in new_blocks[:-1]
    ), "only the last block may be empty"
    assert not any(
        new_cfg.out_edges(new_blocks[-1])
    ), "the last block cannot have outgoing cfg edges"

    bi = block.byte_interval
    assert bi

    _add_return_edges_for_patch_calls(
        block.module, functions_by_block, new_cfg
    )
    _update_patch_return_edges_to_match(
        block, functions_by_block, new_cfg, new_proxy_blocks
    )

    # Adjust codeblock sizes, create CFG edges, remove 0-size blocks
    _modify_block_insert_cfg(
        block,
        offset,
        replacement_length,
        new_bytes,
        new_blocks,
        new_cfg,
        new_symbolic_expressions,
        new_symbols,
        functions_by_block,
    )

    size_delta = len(new_bytes) - replacement_length
    offset += block.offset

    # adjust byte interval the block goes in
    bi.size += size_delta
    bi.contents = (
        bi.contents[:offset]
        + new_bytes
        + bi.contents[offset + replacement_length :]
    )

    # adjust blocks that occur after the insertion point
    # TODO: what if blocks overlap over the insertion point?
    for b in bi.blocks:
        if b != block and b.offset >= offset:
            b.offset += size_delta

    # adjust all of the new blocks to be relative to the byte interval and
    # add them to the byte interval
    for b in new_blocks:
        b.offset += offset

    assert block.size and all(b.size for b in new_blocks), (
        "_modify_block_insert created a zero-sized block; please file a bug "
        "report against gtirb-rewriting"
    )
    bi.blocks.update(new_blocks)

    # adjust sym exprs that occur after the insertion point
    bi.symbolic_expressions = {
        (k + size_delta if k >= offset else k): v
        for k, v in bi.symbolic_expressions.items()
        if k < offset or k >= offset + replacement_length
    }

    # add all of the symbolic expressions from the code we're inserting
    for rel_offset, expr in new_symbolic_expressions.items():
        bi.symbolic_expressions[offset + rel_offset] = expr

    bi.ir.cfg.update(new_cfg)
    bi.module.symbols.update(new_symbols)
    bi.module.proxies.update(new_proxy_blocks)

    # adjust aux data if present
    def update_aux_data_keyed_by_offset(name):
        table = bi.module.aux_data.get(name)
        if table:
            if not isinstance(table.data, OffsetMapping):
                table.data = OffsetMapping(table.data)
            if bi in table.data:
                displacement_map = table.data[bi]
                del table.data[bi]
                table.data[bi] = {
                    (k + size_delta if k >= offset else k): v
                    for k, v in displacement_map.items()
                    if k < offset or k >= offset + replacement_length
                }

    update_aux_data_keyed_by_offset("comments")
    update_aux_data_keyed_by_offset("padding")
    update_aux_data_keyed_by_offset("symbolicExpressionSizes")


def _modify_block_insert_cfg(
    block: gtirb.CodeBlock,
    offset: int,
    replacement_length: int,
    new_bytes: bytes,
    new_blocks: List[gtirb.CodeBlock],
    new_cfg: gtirb.CFG,
    new_symbolic_expressions: Dict[int, gtirb.SymbolicExpression],
    new_symbols: Iterable[gtirb.Symbol],
    functions_by_block: Dict[gtirb.CodeBlock, uuid.UUID],
) -> None:
    """
    Adjust codeblock sizes, create CFG edges, remove 0-size blocks.
    :param block: The code block to insert into.
    :param offset: The byte offset into the code block.
    :param replacement_length: The number of bytes after `offset` that should
                               be removed.
    :param new_bytes: The new content to be inserted.
    :param new_blocks: The blocks corresponding to the new content.
    :param new_cfg: The CFG corresponding to the new content.
    :param new_symbolic_expressions: The symbolic expressions corresponding to
                                     the new content, with the keys relative
                                     to the start of `new_bytes`.
    :param new_symbols: Any symbols that may be defined in the new content.
    :param functions_by_block: Map from code block to containing function UUID.
    """

    bi = block.byte_interval

    original_size = block.size
    inserts_at_end = not replacement_length and offset == block.size
    replaces_last_instruction = (
        replacement_length and offset + replacement_length == block.size
    )

    # Adjust the target codeblock and discard the new codeblock if no new CFG
    # edges are created, no new symbols are created, and the replacement is not
    # at the end of a block.
    if (
        not new_cfg
        and not new_symbols
        and not inserts_at_end
        and not replaces_last_instruction
    ):
        block.size = block.size - replacement_length + len(new_bytes)

        # Remove all the blocks from new_blocks so that they don't get added
        # to the byte_interval in _modify_block_insert.
        new_blocks.clear()
        return

    # Adjust the target block to be the size of offset. Then extend the last
    # patch block to cover the remaining bytes in the original block.
    block.size = offset
    new_blocks[-1].size += original_size - offset - replacement_length

    # Now add a fallthrough edge from the original block to the first patch
    # block, unless we're inserting at the end of the block and the block has
    # no fallthrough edges. For example, inserting after a ret instruction.
    if not inserts_at_end or any(_block_fallthrough_targets(block)):
        new_cfg.add(
            gtirb.Edge(
                source=block,
                target=new_blocks[0],
                label=gtirb.Edge.Label(type=gtirb.Edge.Type.Fallthrough),
            )
        )

    # Alter any outgoing edges from the original block to originate from the
    # last patch block.
    if inserts_at_end:
        fallthrough_targets = _block_fallthrough_targets(block)

        for edge in set(block.outgoing_edges):
            if _is_fallthrough_edge(edge):
                bi.ir.cfg.discard(edge)
                new_cfg.add(edge._replace(source=new_blocks[-1]))
            elif _is_call_edge(edge):
                _update_return_edges_from_changing_fallthrough(
                    edge,
                    fallthrough_targets,
                    functions_by_block,
                    new_blocks[-1],
                    new_cfg,
                )
    elif replaces_last_instruction:
        fallthrough_targets = _block_fallthrough_targets(block)

        for edge in set(block.outgoing_edges):
            if _is_fallthrough_edge(edge):
                bi.ir.cfg.discard(edge)
                new_cfg.add(edge._replace(source=new_blocks[-1]))
            elif _is_call_edge(edge):
                _update_return_edges_from_removing_call(
                    edge, fallthrough_targets, functions_by_block, new_cfg
                )
                bi.ir.cfg.discard(edge)
            else:
                bi.ir.cfg.discard(edge)
    else:
        for edge in set(block.outgoing_edges):
            bi.ir.cfg.discard(edge)
            new_cfg.add(edge._replace(source=new_blocks[-1]))

    # Now go back and clean up any zero-sized blocks, which trigger
    # nondeterministic behavior in the pretty printer.
    if block.size == 0:
        added_fallthrough = next(
            edge
            for edge in new_cfg.in_edges(new_blocks[0])
            if _is_fallthrough_edge(edge) and edge.source == block
        )
        new_cfg.discard(added_fallthrough)

        block.size = new_blocks[0].size
        _substitute_block(
            new_blocks[0], block, new_cfg, new_symbols,
        )
        del new_blocks[0]

    if new_blocks and new_blocks[-1].size == 0:
        has_symbols = any(
            sym.referent == new_blocks[-1] for sym in new_symbols
        )
        has_incoming_edges = any(new_cfg.in_edges(new_blocks[-1]))
        fallthrough_edges = [
            edge
            for edge in new_cfg.out_edges(new_blocks[-1])
            if _is_fallthrough_edge(edge)
        ]

        if not has_symbols and not has_incoming_edges:
            # If nothing refers to the block, we can simply drop it and any
            # outgoing edges that may have been added to it from the earlier
            # steps.
            for out_edge in new_cfg.out_edges(new_blocks[-1]):
                new_cfg.discard(out_edge)
            del new_blocks[-1]
        elif len(fallthrough_edges) == 1:
            # If we know where the "next" block is, substitute that for our
            # last block.
            _substitute_block(
                new_blocks[-1],
                fallthrough_edges[0].target,
                new_cfg,
                new_symbols,
            )
            del new_blocks[-1]
        else:
            # We don't know where control flow goes after our patch, so we'll
            # raise an exception for now. There are other ways of resolving
            # this that we could explore (e.g. insert a nop at the end of the
            # inserted bytes to make it be a non-zero block).
            raise NotImplementedError(
                "Attempting to insert a block at the end of another "
                "block without knowing how to update control flow"
            )


def _is_elf_pie(module: gtirb.Module) -> bool:
    return (
        module.file_format == gtirb.Module.FileFormat.ELF
        and "DYN" in module.aux_data["binaryType"].data
    )


def _text_section_name(module: gtirb.Module):
    if module.file_format == gtirb.Module.FileFormat.ELF:
        return ".text"
    elif module.file_format == gtirb.Module.FileFormat.PE:
        return ".text"
    else:
        assert False, f"unsupported file format: {module.file_format}"


def decorate_extern_symbol(module: gtirb.Module, sym: str) -> str:
    """
    Decorates a symbol as needed for the target. For example, this might
    involve adding a leading underscore on some platforms.
    """
    # TODO: 32-bit Windows uses a leading underscore. So does Mach-O.
    return sym


def effective_alignment(address: int, max_alignment: int = 8) -> int:
    """Return the largest power of two to which an address is aligned."""
    return (~address & (address - 1) & (max_alignment - 1)) + 1


def align_address(address: int, alignment: int) -> int:
    """Increase an address to the next alignment boundary, if necessary."""
    return (address + alignment - 1) & -alignment



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import dataclasses
import itertools
from typing import Iterable, List, Mapping, MutableMapping

import gtirb

from .isa import _get_isa
from .utils import OffsetMapping, align_address, effective_alignment

standard_byte_interval_aux_data = (
    "comments",
    "padding",
    "symbolicExpressionSizes",
)


class PaddingError(Exception):
    """Indicates an error inserting padding to reach a desired alignment."""


@dataclasses.dataclass
class BlockGroup:
    """A group of overlapping blocks."""

    begin: int
    """Offset of the first block in the group."""
    end: int
    """First offset past the end of the last block in the group."""
    blocks: List[gtirb.ByteBlock]
    """Collection of blocks in the group."""


def split_byte_interval(
    interval: gtirb.ByteInterval,
    alignment: MutableMapping[gtirb.Node, int] = None,
    tables: Iterable[OffsetMapping[object]] = None,
) -> List[gtirb.ByteInterval]:
    """Split a ByteInterval to put each block in its own interval.

    After the split, the original interval will hold the first block (ordered
    by offset). Each remaining block will be in a new new byte interval. Any
    bytes outside of a block will be included in the interval containing the
    preceding block; the first interval will contain the bytes before and after
    the first block.

    Because overlapping blocks share the bytes where they overlap, some
    intervals may contain more than one block after the split. These intervals
    will contain the smallest number of blocks possible without duplicating
    bytes.

    :param interval:  byte interval to split
    :param alignment:  optional table of alignments for blocks and intervals
    :param tables:  optional collection of offset mappings to update
    :returns:  list of byte intervals containing the blocks in the original
        interval
    """
    if tables is None:
        tables = []
        module = interval.module
        if module is not None:
            for name in standard_byte_interval_aux_data:
                if name in module.aux_data:
                    aux_data = module.aux_data[name]
                    if not isinstance(aux_data.data, OffsetMapping):
                        aux_data.data = OffsetMapping(aux_data.data)
                    tables.append(aux_data.data)

    # Group overlapping blocks so they can be processed as a unit.
    groups = []
    for block in sorted(interval.blocks, key=lambda b: b.offset):
        block_end = block.offset + block.size
        if groups == [] or groups[-1].end <= block.offset:
            groups.append(BlockGroup(block.offset, block_end, [block]))
        else:
            groups[-1].end = max(groups[-1].end, block_end)
            groups[-1].blocks.append(block)
        if alignment is not None and block not in alignment:
            if block.address is None:
                # Align the offset, since we don't know the actual address
                alignment[block] = effective_alignment(block.offset)
            else:
                alignment[block] = effective_alignment(block.address)

    # Process groups in decreasing offset order, but skip the first group
    # because it will stay in the original interval.
    groups.reverse()
    groups.pop()

    # Create the new byte interval for each group of blocks.
    intervals = []
    offset = interval.size
    for group in groups:
        new_interval = gtirb.ByteInterval(
            contents=interval.contents[group.begin :],
            size=max(offset - group.begin, 0),
        )
        new_interval.section = interval.section

        new_interval.address = group.blocks[0].address
        for block in group.blocks:
            block.offset -= group.begin
            block.byte_interval = new_interval
        intervals.append(new_interval)

        offset = min(interval.size, group.begin)
        interval.initialized_size = min(interval.initialized_size, offset)
        interval.size = min(interval.size, offset)
    intervals.append(interval)

    # Transfer symbolic expressions and table items to the new intervals.
    symexprs = OffsetMapping()
    symexprs[interval] = interval.symbolic_expressions
    for table in itertools.chain((symexprs,), tables):
        items = sorted(table.get(interval, {}).items())
        for group, new_interval in zip(groups, intervals):
            while items != [] and items[-1][0] >= group.begin:
                off, value = items.pop()
                del table[interval][off]
                if new_interval not in table:
                    table[new_interval] = {}
                table[new_interval].update({off - group.begin: value})
    for new_interval in intervals:
        if new_interval in symexprs:
            new_interval.symbolic_expressions = symexprs[new_interval]

    intervals.reverse()
    return intervals


def join_byte_intervals(
    intervals: List[gtirb.ByteInterval],
    nop: bytes = None,
    alignment: Mapping[gtirb.Node, int] = None,
    tables: Iterable[OffsetMapping[object]] = None,
) -> None:
    """Concatenate a list of byte intervals.

    The first interval in the given list will be trated as the destination. The
    contents (bytes and byte_blocks) of all other intervals will be
    concatenated onto the end of the destination interval in the order they
    appear in the list.

    Padding will be inserted between subsequent intervals so that the address
    of the first block of each interval (or of the interval itself if it
    contains no blocks) is properly aligned. Addresses are calculated based on
    the address of the destination block, or 0 if it has no address. If the
    alignment mapping is not specified, the "alignment" aux data for each
    interval's module, if any, will be used.

    The symbolic expressions will be transfered to the destination module,
    adjusted to retain their positions relative to their original byte
    interval. In addition, any tables given will be updated by relocating
    Offsets into each concatenated interval to refer to the corresponding
    Offset into the destination interval. If no tables are provided, a default
    set of aux data will be updated; pass an empty sequence of tables to
    prevent any tables from being updated.

    NB: This function destructively removes the blocks, bytes, and symbolic
    expressions from the other intervals when they are added to the
    destination, but it does not remove the intervals from their sections.

    :param intervals:  list of byte intervals to concatenate
    :param nop:  bytes representing a single nop instruction
    :param alignment:  table of alignments for blocks and intervals
    :param tables:  collection of offset mappings to update
    """
    if len(intervals) < 2:
        return

    if tables is None:
        # This is a bit hacky, but to avoid assuming that the byte intervals
        # are all in the same module, the tables are a list of dictionaries
        # that map intervals to (displacement to value) dicts. Each interval
        # added at this stage will map to the mutable mapping returned by
        # indexing an OffsetMapping, which means the original aux data will be
        # updated when modifying that sub-dict.
        tables = []
        for name in standard_byte_interval_aux_data:
            table = {}
            for bi in intervals:
                if bi.module is not None and name in bi.module.aux_data:
                    aux_data = bi.module.aux_data[name]
                    if not isinstance(aux_data.data, OffsetMapping):
                        aux_data.data = OffsetMapping(aux_data.data)
                    if bi in aux_data.data:
                        table[bi] = aux_data.data[bi]
            if len(table) > 0:
                tables.append(table)

    destination = intervals.pop(0)

    address = 0
    if destination.address is not None:
        address = destination.address
    address += destination.size
    last_block = max(destination.blocks, key=lambda b: b.offset, default=None)
    last_module = last_block.module if last_block is not None else None

    def insert_padding(size):
        if size == 0:
            return
        if isinstance(last_block, gtirb.CodeBlock):
            BlockType = gtirb.CodeBlock
            if nop is not None:
                pad_bytes = nop
            elif last_module is not None:
                pad_bytes = _get_isa(last_module).nop()
            else:
                raise PaddingError("cannot determine nop instruction")
            size, remainder = divmod(size, len(pad_bytes))
            if remainder != 0:
                raise PaddingError("nop does not fit evenly in padding")
        else:
            BlockType = gtirb.DataBlock
            pad_bytes = b"\x00"

        # The pretty-printer won't print the padding bytes unless they
        # are contained in blocks.
        padding = BlockType(offset=len(destination.contents), size=size)
        padding.byte_interval = destination
        destination.contents += pad_bytes * size

    symexprs = OffsetMapping()
    deltas = {}
    for interval in intervals:
        # Fill in any uninitialized bytes before appending.
        insert_padding(destination.size - len(destination.contents))

        # Align the first block if possible, or the interval if not.
        if alignment is not None:
            module_alignment = alignment
        elif (
            interval.module is not None
            and "alignment" in interval.module.aux_data
        ):
            module_alignment = interval.module.aux_data["alignment"].data
        else:
            module_alignment = {}
        node = min(
            (b for b in interval.blocks if b in module_alignment),
            key=lambda b: b.offset,
            default=interval,
        )
        if node == interval:
            offset = 0
        else:
            offset = node.offset
        boundary = module_alignment.get(node, 1)
        size = align_address(address + offset, boundary) - (address + offset)
        insert_padding(size)
        address += size
        destination.size += size

        # Cache the delta for updating the symbolic expression offsets and the
        # new last block in case we need more padding.
        deltas[interval] = len(destination.contents)
        symexprs[interval] = interval.symbolic_expressions
        last_block = max(
            interval.blocks, default=last_block, key=lambda b: b.offset
        )
        if last_block is not None and last_block.module is not None:
            last_module = last_block.module

        # Transfer the bytes and blocks to the new intervals.
        address += interval.size
        destination.size += interval.size
        destination.contents += interval.contents
        for block in tuple(interval.blocks):
            block.offset += deltas[interval]
            block.byte_interval = destination

        interval.initialized_size = 0
        interval.symbolic_expressions.clear()

    destination.initialized_size = len(destination.contents)

    # Update offsets to refer to the destination interval.
    for table in itertools.chain((symexprs,), tables):
        for interval in intervals:
            old = table.get(interval, {})
            new_items = ((k + deltas[interval], v) for k, v in old.items())
            if destination in table:
                table[destination].update(new_items)
            else:
                table[destination] = dict(new_items)
            old.clear()
    destination.symbolic_expressions.update(symexprs[destination])

    return destination



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
__version__ = "0.0.11"



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import imp

import setuptools

version = imp.load_source(
    "pkginfo.version", "gtirb_rewriting/version.py"
).__version__

setuptools.setup(
    name="gtirb-rewriting",
    version=version,
    author="GrammaTech",
    author_email="gtirb@grammatech.com",
    description="Utilities for rewriting GTIRB",
    packages=setuptools.find_packages(),
    install_requires=[
        "gtirb-capstone",
        "gtirb-functions",
        "gtirb",
        "mcasm",
        "dataclasses ; python_version<'3.7.0'",
    ],
    classifiers=["Programming Language :: Python :: 3"],
    extras_require={
        "test": [
            "flake8",
            "isort",
            "pytest",
            "pytest-cov",
            "tox",
            "tox-wheel",
            "pre-commit",
        ]
    },
    url="https://github.com/grammatech/gtirb-rewriting",
    license="GPLv3",
)



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
from .assembly import (
    Constraints,
    InsertionContext,
    Patch,
    Register,
    X86Syntax,
    patch_constraints,
)
from .intervalutils import (
    PaddingError,
    join_byte_intervals,
    split_byte_interval,
)
from .passes import Pass, PassManager
from .rewriting import RewritingContext
from .scopes import (
    ENTRYPOINT_NAME,
    MAIN_NAME,
    AllBlocksScope,
    AllFunctionsScope,
    BlockPosition,
    FunctionPosition,
    Scope,
    SingleBlockScope,
    pattern_match,
)
from .utils import OffsetMapping, decorate_extern_symbol, show_block_asm
from .version import __version__

__all__ = [
    "__version__",
    "AllBlocksScope",
    "AllFunctionsScope",
    "BlockPosition",
    "Constraints",
    "decorate_extern_symbol",
    "ENTRYPOINT_NAME",
    "FunctionPosition",
    "InsertionContext",
    "join_byte_intervals",
    "MAIN_NAME",
    "OffsetMapping",
    "PaddingError",
    "Pass",
    "PassManager",
    "patch_constraints",
    "pattern_match",
    "Patch",
    "Register",
    "RewritingContext",
    "Scope",
    "SingleBlockScope",
    "show_block_asm",
    "split_byte_interval",
    "X86Syntax",
]



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import contextlib
from typing import Iterator

import gtirb

from .intervalutils import join_byte_intervals, split_byte_interval


@contextlib.contextmanager
def prepare_for_rewriting(module: gtirb.Module, nop: bytes) -> Iterator[None]:
    """Pre-compute data structure to accelerate rewriting."""

    alignment = (
        {} if module.file_format == gtirb.Module.FileFormat.ELF else None
    )
    if "alignment" in module.aux_data:
        alignment = module.aux_data["alignment"].data

    partitions = []
    for interval in tuple(module.byte_intervals):
        if any(isinstance(b, gtirb.CodeBlock) for b in interval.blocks):
            partitions.append(split_byte_interval(interval, alignment))

    yield

    for partition in partitions:
        join_byte_intervals(partition, nop, alignment)
        for interval in partition[1:]:
            interval.section = None



# GTIRB-Rewriting Rewriting API for GTIRB
# Copyright (C) 2021 GrammaTech, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# This project is sponsored by the Office of Naval Research, One Liberty
# Center, 875 N. Randolph Street, Arlington, VA 22203 under contract #
# N68335-17-C-0700.  The content of the information does not necessarily
# reflect the position or policy of the Government and no official
# endorsement should be inferred.
import logging
from typing import Sequence

import gtirb
import gtirb_functions

from .rewriting import RewritingContext


class Pass:
    def begin_module(
        self,
        module: gtirb.Module,
        functions: Sequence[gtirb_functions.Function],
        rewriting_ctx: RewritingContext,
    ) -> None:
        """
        Invoked by the pass manager at the beginning of rewriting a module.
        Passes should use the rewriting context to register their insertions
        in this callback.

        :param module: The module being rewritten.
        :param functions: The calculated list of functions in the module.
        :param rewriting_context: The context to register modifications with.
        """
        pass

    def end_module(
        self,
        module: gtirb.Module,
        functions: Sequence[gtirb_functions.Function],
    ) -> None:
        """
        Invoked by the pass manager after applying rewrites to a module.

        :param module: The module being rewritten.
        :param functions: The calculated list of functions in the module.
        """
        pass


class PassManager:
    """
    Maintains a list of registered passes and runs them on IR.
    """

    def __init__(
        self,
        logger=logging.getLogger("gtirb_rewriting"),
        expensive_assertions=True,
    ):
        """
        :param logger: The logger to log to when rewriting.
        :param expensive_assertions: If enabled, extra assertions will be
        enabled that may have noticable run-time overhead.
        """
        self._logger = logger
        self._passes = []
        self._expensive_assertions = expensive_assertions

    def add(self, pass_inst: Pass) -> None:
        """
        Registers a pass with the pass manager.
        """
        self._passes.append(pass_inst)

    def run(self, ir: gtirb.IR) -> None:
        """
        Runs the passes on the GTIRB IR.
        """
        for mod in ir.modules:
            functions = gtirb_functions.Function.build_functions(mod)
            context = RewritingContext(
                mod,
                functions,
                logger=self._logger,
                expensive_assertions=self._expensive_assertions,
            )

            for pass_inst in self._passes:
                pass_inst.begin_module(mod, functions, context)

            context.apply()

            for pass_inst in self._passes:
                pass_inst.end_module(mod, functions)
